/*
 * Copyright (C) 2009 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

// OpenGL ES 2.0 code

#include <jni.h>
#include <android/log.h>

#include <GLES2/gl2.h>
#include <GLES2/gl2ext.h>

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <sys/time.h>

#define  LOG_TAG    "libgl2jni"
#define  LOGI(...)  __android_log_print(ANDROID_LOG_INFO,LOG_TAG,__VA_ARGS__)
#define  LOGE(...)  __android_log_print(ANDROID_LOG_ERROR,LOG_TAG,__VA_ARGS__)

long long getTime()
{
    struct timeval tv;
    gettimeofday(&tv, NULL);
    return ((tv.tv_sec * 1000) + (tv.tv_usec / 1000));
}

auto gVertexShader =
    "attribute vec2 vPosition;\n"
    "attribute vec2 gTexUV;\n"
    "uniform vec2 mat;\n"
    "varying vec2 uv;\n"
    "varying vec2 pos;\n"
    "void main() {\n"
    "  pos = mat+vPosition;\n"
    "  gl_Position = vec4(pos,0.0,1.0);\n"
    "  uv = gTexUV;\n"
    "}\n";

auto gFragmentShader =
    "precision mediump float;\n"
    "uniform sampler2D tex;\n"
    "uniform vec2 ratios;\n"
    "uniform float wave_radius;\n"
    "uniform vec2 wave_center;\n"
    "varying vec2 uv;\n"
    "varying vec2 pos;\n"
    "void main() {\n"
    "  float dist = length((pos-wave_center)*ratios);\n"
    "  if (dist < wave_radius) {\n"
    "    if ( (dist > (wave_radius-0.01)) "
    "     || (dist > (wave_radius-0.04) && dist < (wave_radius-0.035))"
    "     || (dist > (wave_radius-0.14) && dist < (wave_radius-0.135))"
    "     || (dist > (wave_radius-0.34) && dist < (wave_radius-0.335)) ) {\n"
    "      gl_FragColor = vec4(1.0, 1.0, 1.0, 1.0);\n"
    "    } else {\n"
    "      float factor = 1.0-(dist/wave_radius)*0.5;\n"
    "      gl_FragColor = texture2D(tex,uv)*vec4(factor,factor,factor,1.0);\n"
    "    }\n"
    "  } else {\n"
    "    gl_FragColor = vec4(0.0, 0.0, 0.0, 1.0);\n"
    "  }\n"
    "}\n";

#define rand getTime

GLuint loadShader(GLenum shaderType, const char* pSource) {
    GLuint shader = glCreateShader(shaderType);
    if (shader) {
        glShaderSource(shader, 1, &pSource, NULL);
        glCompileShader(shader);
        GLint compiled = 0;
        glGetShaderiv(shader, GL_COMPILE_STATUS, &compiled);
        if (!compiled) {
            GLint infoLen = 0;
            glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &infoLen);
            if (infoLen) {
                char* buf = (char*) malloc(infoLen);
                if (buf) {
                    glGetShaderInfoLog(shader, infoLen, NULL, buf);
                    LOGE("Could not compile shader %d:\n%s\n",shaderType, buf);
                    free(buf);
                }
                glDeleteShader(shader);
                shader = 0;
            }
        }
    }
    return shader;
}

GLuint createProgram(const char* pVertexSource, const char* pFragmentSource) {
    GLuint vertexShader = loadShader(GL_VERTEX_SHADER, pVertexSource);
    if (!vertexShader) {
        return 0;
    }

    GLuint pixelShader = loadShader(GL_FRAGMENT_SHADER, pFragmentSource);
    if (!pixelShader) {
        return 0;
    }

    GLuint program = glCreateProgram();
    if (program) {
        glAttachShader(program, vertexShader);
        glAttachShader(program, pixelShader);
        glLinkProgram(program);
        GLint linkStatus = GL_FALSE;
        glGetProgramiv(program, GL_LINK_STATUS, &linkStatus);
        if (linkStatus != GL_TRUE) {
            GLint bufLength = 0;
            glGetProgramiv(program, GL_INFO_LOG_LENGTH, &bufLength);
            if (bufLength) {
                char* buf = (char*) malloc(bufLength);
                if (buf) {
                    glGetProgramInfoLog(program, bufLength, NULL, buf);
                    LOGE("Could not link program:\n%s\n", buf);
                    free(buf);
                }
            }
            glDeleteProgram(program);
            program = 0;
        }
    }
    return program;
}

GLfloat gObstacleVertices[] = { +0.1f, +0.1f,
                                +0.1f, -0.1f,
                                -0.1f, +0.1f,
                                +0.1f, -0.1f,
                                -0.1f, +0.1f,
                                -0.1f, -0.1f };

const GLfloat gObstacleUV[] = { 1.0f, 0.0f,
                                1.0f, 1.0f,
                                0.0f, 0.0f,
                                1.0f, 1.0f,
                                0.0f, 0.0f,
                                0.0f, 1.0f };

GLuint gProgram;
GLuint gvPositionHandle;
GLuint gvUVHandle;
GLuint gvMatHandle;
GLuint gvTexHandle;
GLuint gvWaveCenterHandle;
GLuint gvWaveRadiusHandle;
GLuint gvRatiosHandle;

int screen_width, screen_height;
int screen_width_offset, screen_height_offset;

float min_x;
float min_y;
float cell_x;
float cell_y;
float height_ratio, width_ratio;

float pause_button_x;
float pause_button_y;

float level_icon_x;
float level_icon_y;

float start_button_x;
float start_button_y;
float start_button_width;
float start_button_height;

enum GameScreen
{
    MAIN_MENU,
    GAME_PLAY,
    CREDITS_SCREEN,
    TUTORIAL_SCREEN,
};

GameScreen game_screen = MAIN_MENU;

enum GameObject
{
    TEXT,
    WALL,
    BAT_BACK_1,
    BAT_BACK_2,
    BAT_BACK_3,
    BAT_FRONT_1,
    BAT_FRONT_2,
    BAT_FRONT_3,
    BAT_LEFT_1,
    BAT_LEFT_2,
    BAT_RIGHT_1,
    BAT_RIGHT_2,
    EATER_1,
    EATER_2,
    ARROW_LEFT,
    ARROW_RIGHT,
    ARROW_UP,
    ARROW_DOWN,
    CALM_GHOST_LEFT,
    CALM_GHOST_RIGHT,
    ANGRY_GHOST_LEFT,
    ANGRY_GHOST_RIGHT,
    GROUND,
    PAUSE,
    PLAY,
    ENTERANCE,
    EXIT,
    EXIT_OPEN,
    DEAD,
    PAUSED,
    START,
    OPENING,
    NICE,
    KEY,
    LV1,
    LV2,
    LV3,
    TUTORIAL,
    TUTORIAL_CONTENTS,
    CREDITS,
    CREDITS_CONTENTS,
};

typedef int GameMap[10][10];

struct Ghost
{
    int ghost_x = 4;
    int ghost_y = 9;
    int ghost_direction_x = -1;
    int ghost_direction_y = -1;
    int ghost_goal_x;
    int ghost_goal_y;
    bool ghost_goal_detected = false;
    int ghost_cycle;
    GameObject ghost_frame;
};

struct Level
{
    int num_ghosts;
    Ghost init_ghosts[10];
    bool has_key;
    float wave_range;
    GameMap* map;
    int rows;
    int columns;
    int goal_x;
    int goal_y;
    char* message;
};

#define NUM_LEVELS 3

Level all_levels[NUM_LEVELS];

Ghost current_ghosts[10];

GameMap current_map;

int current_level = 0;

float touch_x, touch_y;
bool touch_event = false;

bool paused = false;
bool level_intro = true;
float level_intro_pos = 1.2f;

struct TextureInfo
{
    int width;
    int height;
    char* data;
    GLuint textureID;
    GLenum format;
};

static TextureInfo gTextureList[30];
static int gNumTextureList = 0;

int bat_x = 0;
int bat_y = 0;

bool diededed = false;
int diededed_counter = 0;
int killer_x, killer_y;

bool killed_by_monster = false;
bool killed_by_ghost = false;

bool reached_goal = false;
bool found_keys = false;
bool just_found_keys = false;
int reached_goal_count = 0;

int last_bat_movement = 1;

GameObject character_frame = BAT_FRONT_1;

float wave_rad = 0.0f;

bool fire_photon_torpedos = false;
bool stable_photon_torpedos = false;
int stablizing_counter = 0;

bool landscape;

bool setupGraphics(int w, int h) {

    screen_height = h;
    screen_width = w;

    if (h > w)
    {
        int extra = h-w;
        screen_height_offset = extra/2;
        screen_width_offset = 0;
        float ratio = w/(float)h;
        cell_x = 0.2f;
        cell_y = 0.2f*ratio;
        min_x = -0.9f;
        min_y = cell_y*0.5f-ratio;
        width_ratio = 1.0f;
        height_ratio = ratio;

        pause_button_x = 0.0f;
        pause_button_y = 0.85f;

        start_button_x = 0.0f;
        start_button_y = 0.0f;

        start_button_width = 0.4f;
        start_button_height = 0.2f;
        landscape = false;
    }
    else
    {
        int extra = w-h;
        screen_width_offset = extra/2;
        screen_height_offset = 0;
        float ratio = h/(float)w;
        cell_y = 0.2f;
        cell_x = 0.2f*ratio;
        min_y = -0.9f;
        min_x = cell_y*0.5f-ratio;
        height_ratio = 1.0f;
        width_ratio = ratio;

        pause_button_x = -0.85f;
        pause_button_y = 0.0f;

        start_button_x = 0.0f;
        start_button_y = 0.0f;

        start_button_height = 0.4f;
        start_button_width = 0.2f;
        landscape = true;
    }

    GLfloat cell_verts[] = { +cell_x/2, +cell_y/2,
                             +cell_x/2, -cell_y/2,
                             -cell_x/2, +cell_y/2,
                             +cell_x/2, -cell_y/2,
                             -cell_x/2, +cell_y/2,
                             -cell_x/2, -cell_y/2 };

    memcpy(gObstacleVertices, cell_verts, sizeof(gObstacleVertices));

    gProgram = createProgram(gVertexShader, gFragmentShader);
    if (!gProgram) {
        return false;
    }
    gvPositionHandle = glGetAttribLocation(gProgram, "vPosition");
    gvUVHandle = glGetAttribLocation(gProgram, "gTexUV");
    gvMatHandle = glGetUniformLocation(gProgram, "mat");
    gvTexHandle = glGetUniformLocation(gProgram, "tex");
    gvWaveCenterHandle = glGetUniformLocation(gProgram, "wave_center");
    gvWaveRadiusHandle = glGetUniformLocation(gProgram, "wave_radius");
    gvRatiosHandle = glGetUniformLocation(gProgram, "ratios");

    for (int i = 0 ; i < gNumTextureList ; i++)
    {
        glGenTextures(1, &gTextureList[i].textureID);
        glBindTexture(GL_TEXTURE_2D, gTextureList[i].textureID);
        glTexImage2D(GL_TEXTURE_2D, 0, gTextureList[i].format, gTextureList[i].width, gTextureList[i].height, 0, gTextureList[i].format, GL_UNSIGNED_BYTE, gTextureList[i].data);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
        glGenerateMipmap(GL_TEXTURE_2D);
    }

    glViewport(0, 0, w, h);

    return true;
}

void InitializeLevels()
{
    static bool initialized = false;
    if (!initialized)
    {
        /* Initialize Level 1 */
        {
            static GameMap game_map_1 = {
                    { 1, 1, 1, 1, 1, 1, 1, 1, 1, 0 },
                    { 1, 2, 0, 0, 0, 1, 0, 0, 0, 0 },
                    { 1, 1, 1, 2, 0, 1, 0, 1, 1, 1 },
                    { 0, 0, 0, 0, 0, 1, 0, 0, 0, 0 },
                    { 0, 2, 1, 1, 1, 1, 0, 2, 0, 2 },
                    { 0, 1, 1, 0, 0, 0, 0, 0, 0, 0 },
                    { 0, 1, 0, 0, 1, 1, 1, 1, 1, 0 },
                    { 0, 1, 0, 0, 1, 0, 2, 0, 1, 0 },
                    { 0, 1, 1, 1, 1, 0, 0, 0, 1, 0 },
                    { 0, 0, 0, 0, 0, 0, 2, 0, 0, 0 },
            };

            static char* message_1 = "Becareful of the things in the dark!";

            all_levels[0].message = message_1;
            all_levels[0].map = &game_map_1;
            all_levels[0].columns = 10;
            all_levels[0].rows = 10;
            all_levels[0].goal_x = 9;
            all_levels[0].goal_y = 9;
            all_levels[0].has_key = false;
            all_levels[0].wave_range = 0.75f;
            all_levels[0].num_ghosts = 0;
        }

        /* Initialize Level 2 */
        {
            static GameMap game_map_2 = {
                    { 1, 1, 1, 1, 1, 1, 1, 0, 0, 0 },
                    { 0, 0, 0, 0, 2, 1, 2, 0, 1, 1 },
                    { 0, 1, 1, 0, 0, 0, 0, 0, 1, 0 },
                    { 0, 1, 0, 1, 1, 1, 1, 1, 1, 0 },
                    { 0, 0, 2, 1, 0, 0, 0, 1, 0, 0 },
                    { 0, 1, 1, 1, 0, 1, 2, 1, 0, 1 },
                    { 0, 0, 0, 2, 0, 1, 1, 1, 0, 1 },
                    { 1, 1, 0, 0, 0, 0, 0, 0, 0, 0 },
                    { 0, 1, 0, 1, 1, 1, 1, 0, 1, 2 },
                    { 0, 0, 0, 0, 0, 2, 1, 0, 1, 0 },
            };

            static char* message_2 = "Your radar might annoy the ghost!";

            all_levels[1].message = message_2;
            all_levels[1].map = &game_map_2;
            all_levels[1].columns = 10;
            all_levels[1].rows = 10;
            all_levels[1].goal_x = 9;
            all_levels[1].goal_y = 9;
            all_levels[1].has_key = false;
            all_levels[1].wave_range = 0.5f;
            all_levels[1].num_ghosts = 1;
            all_levels[1].init_ghosts[0].ghost_cycle = 30;
            all_levels[1].init_ghosts[0].ghost_direction_x = 1;
            all_levels[1].init_ghosts[0].ghost_direction_y = 1;
            all_levels[1].init_ghosts[0].ghost_frame = CALM_GHOST_LEFT;
            all_levels[1].init_ghosts[0].ghost_goal_detected = false;
            all_levels[1].init_ghosts[0].ghost_x = 4;
            all_levels[1].init_ghosts[0].ghost_y = 9;
        }

        /* Initialize Level 3 */
        {
            static GameMap game_map_3 = {
                    { 1, 1, 1, 1, 1, 1, 1, 1, 1, 0 },
                    { 1, 2, 3, 0, 1, 0, 1, 0, 0, 0 },
                    { 1, 1, 1, 0, 0, 0, 1, 0, 0, 0 },
                    { 1, 2, 0, 0, 1, 0, 1, 1, 2, 0 },
                    { 1, 0, 0, 0, 1, 0, 2, 0, 0, 0 },
                    { 1, 1, 2, 0, 1, 0, 1, 0, 1, 1 },
                    { 0, 0, 0, 0, 1, 0, 2, 0, 0, 0 },
                    { 0, 1, 1, 1, 1, 0, 1, 1, 1, 0 },
                    { 0, 1, 2, 1, 1, 0, 0, 0, 0, 0 },
                    { 0, 0, 0, 2, 1, 2, 1, 1, 1, 0 },
            };

            static char* message_3 = "Find the key to open the door!";

            all_levels[2].message = message_3;
            all_levels[2].map = &game_map_3;
            all_levels[2].columns = 10;
            all_levels[2].rows = 10;
            all_levels[2].goal_x = 9;
            all_levels[2].goal_y = 9;
            all_levels[2].has_key = true;
            all_levels[2].wave_range = 0.5f;
            all_levels[2].num_ghosts = 2;
            all_levels[2].init_ghosts[0].ghost_cycle = 35;
            all_levels[2].init_ghosts[0].ghost_direction_x = 1;
            all_levels[2].init_ghosts[0].ghost_direction_y = 1;
            all_levels[2].init_ghosts[0].ghost_frame = CALM_GHOST_LEFT;
            all_levels[2].init_ghosts[0].ghost_goal_detected = false;
            all_levels[2].init_ghosts[0].ghost_x = 4;
            all_levels[2].init_ghosts[0].ghost_y = 9;
            all_levels[2].init_ghosts[1].ghost_cycle = 35;
            all_levels[2].init_ghosts[1].ghost_direction_x = 1;
            all_levels[2].init_ghosts[1].ghost_direction_y = 1;
            all_levels[2].init_ghosts[1].ghost_frame = CALM_GHOST_LEFT;
            all_levels[2].init_ghosts[1].ghost_goal_detected = false;
            all_levels[2].init_ghosts[1].ghost_x = 6;
            all_levels[2].init_ghosts[1].ghost_y = 6;
        }

        initialized = true;
    }
}

bool getTouchedCell(int& x, int& y)
{
    if (touch_event)
    {
        float relative_x = touch_x-screen_width_offset;
        float relative_y = touch_y-screen_height_offset;
        int limited_width = screen_width-screen_width_offset*2;
        int limited_height = screen_height-screen_height_offset*2;
        x = int((relative_x/limited_width)*10);
        y = 9-int((relative_y/limited_height)*10);

        if (x < 0 || x > 9 || y < 0 || y > 9)
        {
            x = 0;
            y = 0;
            return false;
        }
        else
        {
            return true;
        }
    }
    else
    {
        return false;
    }
}

bool isPausePlayTouched()
{
    if (touch_event)
    {
        float x = (touch_x/(float)screen_width)*2.0f-1.0f;
        float y = 1.0f-(touch_y/(float)screen_height)*2.0f;
        if (fabs(x-pause_button_x) < 0.1f && fabs(y-pause_button_y) < 0.1f)
        {
            return true;
        }
    }
    return false;
}

bool isStartTouched()
{
    if (touch_event)
    {
        float x = (touch_x/(float)screen_width)*2.0f-1.0f;
        float y = 1.0f-(touch_y/(float)screen_height)*2.0f;
        if (fabs(x-start_button_x) < start_button_width*0.5f && fabs(y-start_button_y) < start_button_height*0.5f)
        {
            return true;
        }
    }
    return false;
}

bool isCreditsTouched()
{
    if (touch_event)
    {
        float x = (touch_x/(float)screen_width)*2.0f-1.0f;
        float y = 1.0f-(touch_y/(float)screen_height)*2.0f;
        if (fabs(x-(0.5)) < 0.3f && fabs(y-(-0.4)) < 0.5f)
        {
            return true;
        }
    }
    return false;
}

bool isTutorialTouched()
{
    if (touch_event)
    {
        float x = (touch_x/(float)screen_width)*2.0f-1.0f;
        float y = 1.0f-(touch_y/(float)screen_height)*2.0f;
        if (fabs(x-(-0.5)) < 0.3f && fabs(y-(-0.4)) < 0.5f)
        {
            return true;
        }
    }
    return false;
}

void WriteText(const char* text, float x, float y, float scale, bool center = true)
{
    const float char_width = 1.0f/16.0f;
    const float char_height = 134.0f/1024.0f;

    size_t len = strlen(text);

    float start_x = x;

    if (center)
    {
        start_x = x - scale*len + scale;
    }

    glUseProgram(gProgram);

    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D,gTextureList[TEXT].textureID);
    glUniform1i(gvTexHandle, 0);

    for (int i = 0 ; i < len ; i++)
    {
        char c = text[i];
        if (c >= 32 && c <= 127)
        {
            c -= 32;

            float uv_x_start = (c%16)*char_width;
            float uv_x_end = uv_x_start+char_width;
            float uv_y_start = (c/16)*char_height;
            float uv_y_end = uv_y_start+char_height;

            GLfloat uv[] = { uv_x_end,   uv_y_start,
                             uv_x_end,   uv_y_end,
                             uv_x_start, uv_y_start,
                             uv_x_end,   uv_y_end,
                             uv_x_start, uv_y_start,
                             uv_x_start, uv_y_end };

//            GLfloat verts[] = { +scale*width_ratio, +scale*height_ratio,
//                                +scale*width_ratio, -scale*height_ratio,
//                                -scale*width_ratio, +scale*height_ratio,
//                                +scale*width_ratio, -scale*height_ratio,
//                                -scale*width_ratio, +scale*height_ratio,
//                                -scale*width_ratio, -scale*height_ratio };

            GLfloat verts[] = { +scale, +scale,
                                +scale, -scale,
                                -scale, +scale,
                                +scale, -scale,
                                -scale, +scale,
                                -scale, -scale };

            glEnable(GL_BLEND);
            glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
            glUseProgram(gProgram);
            glVertexAttribPointer(gvPositionHandle, 2, GL_FLOAT, GL_FALSE, 0, verts);
            glEnableVertexAttribArray(gvPositionHandle);
            glVertexAttribPointer(gvUVHandle, 2, GL_FLOAT, GL_FALSE, 0, uv);
            glEnableVertexAttribArray(gvUVHandle);
            glUniform1i(gvTexHandle, 0);
            glUniform1f(gvWaveRadiusHandle, 10.0f);
            glUniform2f(gvRatiosHandle, 1.0f/width_ratio, 1.0f/height_ratio);
            glUniform2f(gvMatHandle, start_x+i*scale*2, y);
            glDrawArrays(GL_TRIANGLES, 0, 6);
        }
    }
}

void drawObject(GameObject obj, int x, int y, float wave_rad)
{
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
    glUseProgram(gProgram);
    glVertexAttribPointer(gvPositionHandle, 2, GL_FLOAT, GL_FALSE, 0, gObstacleVertices);
    glEnableVertexAttribArray(gvPositionHandle);
    glVertexAttribPointer(gvUVHandle, 2, GL_FLOAT, GL_FALSE, 0, gObstacleUV);
    glEnableVertexAttribArray(gvUVHandle);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D,gTextureList[obj].textureID);
    glUniform1i(gvTexHandle, 0);
    glUniform2f(gvWaveCenterHandle, min_x+bat_x*cell_x, min_y+bat_y*cell_y);
    glUniform1f(gvWaveRadiusHandle, wave_rad);
    glUniform2f(gvRatiosHandle, 1.0f/width_ratio, 1.0f/height_ratio);
    glUniform2f(gvMatHandle, min_x+x*cell_x, min_y+y*cell_y);
    glDrawArrays(GL_TRIANGLES, 0, 6);
}

void drawObject(GameObject obj, float x, float y)
{
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
    glUseProgram(gProgram);
    glVertexAttribPointer(gvPositionHandle, 2, GL_FLOAT, GL_FALSE, 0, gObstacleVertices);
    glEnableVertexAttribArray(gvPositionHandle);
    glVertexAttribPointer(gvUVHandle, 2, GL_FLOAT, GL_FALSE, 0, gObstacleUV);
    glEnableVertexAttribArray(gvUVHandle);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D,gTextureList[obj].textureID);
    glUniform1i(gvTexHandle, 0);
    glUniform1f(gvWaveRadiusHandle, 10.0f);
    glUniform2f(gvRatiosHandle, 1.0f/width_ratio, 1.0f/height_ratio);
    glUniform2f(gvMatHandle, x, y);
    glDrawArrays(GL_TRIANGLES, 0, 6);
}

void drawObjectWithScale(GameObject obj, float x, float y, float scale)
{
    GLfloat verts[] = { +scale*width_ratio, +scale*height_ratio,
                        +scale*width_ratio, -scale*height_ratio,
                        -scale*width_ratio, +scale*height_ratio,
                        +scale*width_ratio, -scale*height_ratio,
                        -scale*width_ratio, +scale*height_ratio,
                        -scale*width_ratio, -scale*height_ratio };

    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
    glUseProgram(gProgram);
    glVertexAttribPointer(gvPositionHandle, 2, GL_FLOAT, GL_FALSE, 0, verts);
    glEnableVertexAttribArray(gvPositionHandle);
    glVertexAttribPointer(gvUVHandle, 2, GL_FLOAT, GL_FALSE, 0, gObstacleUV);
    glEnableVertexAttribArray(gvUVHandle);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D,gTextureList[obj].textureID);
    glUniform1i(gvTexHandle, 0);
    glUniform1f(gvWaveRadiusHandle, 10.0f);
    glUniform2f(gvRatiosHandle, 1.0f/width_ratio, 1.0f/height_ratio);
    glUniform2f(gvMatHandle, x, y);
    glDrawArrays(GL_TRIANGLES, 0, 6);
}

int map(int x, int y)
{
    return current_map[9-y][x];
}

void takeKey(int x, int y)
{
    current_map[9-y][x] = 0;
}

void updateGhostPosition(Ghost* ghost, bool randomize)
{
    bool goal_reached = (ghost->ghost_goal_x == ghost->ghost_x) && (ghost->ghost_goal_y == ghost->ghost_y);
    bool character_still_there = (bat_x == ghost->ghost_goal_x) && (bat_y == ghost->ghost_goal_y);
    if (goal_reached && !character_still_there)
    {
        ghost->ghost_goal_detected = false;
    }

    if (!ghost->ghost_goal_detected)
    {
        if (ghost->ghost_x <= 0)
        {
            ghost->ghost_direction_x = 1;
        }
        else if (ghost->ghost_x >= 9)
        {
            ghost->ghost_direction_x = -1;
        }
        else if (randomize)
        {
            if (rand()%2) ghost->ghost_direction_x *= -1;
        }

        if (ghost->ghost_direction_x == -1)
        {
            ghost->ghost_frame = CALM_GHOST_RIGHT;
        }
        else
        {
            ghost->ghost_frame = CALM_GHOST_LEFT;
        }

        if (ghost->ghost_y <= 0)
        {
            ghost->ghost_direction_y = 1;
        }
        else if (ghost->ghost_y >= 9)
        {
            ghost->ghost_direction_y = -1;
        }
        else if (randomize)
        {
            if (rand()%2) ghost->ghost_direction_y *= -1;
        }

        ghost->ghost_x += ghost->ghost_direction_x;
        ghost->ghost_y += ghost->ghost_direction_y;
    }
    else
    {
        ghost->ghost_frame = ANGRY_GHOST_LEFT;
        if (ghost->ghost_x > ghost->ghost_goal_x)
        {
            ghost->ghost_x--;
            ghost->ghost_frame = ANGRY_GHOST_LEFT;
        }
        else if (ghost->ghost_x < ghost->ghost_goal_x)
        {
            ghost->ghost_x++;
            ghost->ghost_frame = ANGRY_GHOST_RIGHT;
        }
        if (ghost->ghost_y > ghost->ghost_goal_y)
        {
            ghost->ghost_y--;
        }
        else if (ghost->ghost_y < ghost->ghost_goal_y)
        {
            ghost->ghost_y++;
        }
    }
}

void updateCharacterFrame(int last_movement)
{
    static int frame_num = 0;
    switch (last_movement)
    {
        case 0:
            frame_num = (frame_num+1)%3;
            character_frame = (enum GameObject)(BAT_BACK_1+frame_num);
            break;
        case 1:
            frame_num = (frame_num+1)%3;
            character_frame = (enum GameObject)(BAT_FRONT_1+frame_num);
            break;
        case 2:
            frame_num = (frame_num+1)%2;
            character_frame = (enum GameObject)(BAT_LEFT_1+frame_num);
            break;
        case 3:
            frame_num = (frame_num+1)%2;
            character_frame = (enum GameObject)(BAT_RIGHT_1+frame_num);
            break;
    }
}

void restartGame()
{
    reached_goal = false;
    reached_goal_count = 0;
    fire_photon_torpedos = false;
    stable_photon_torpedos = false;
    stablizing_counter = 0;
    wave_rad = 0.0f;
    bat_x = 0;
    bat_y = 0;
    last_bat_movement = 1;
    character_frame = BAT_FRONT_2;
    diededed = false;
    diededed_counter = 0;
    killer_x = -1;
    killer_y = -1;
    killed_by_monster = false;
    killed_by_ghost = false;
    memcpy(current_ghosts,all_levels[current_level].init_ghosts,sizeof(current_ghosts));
    memcpy(current_map,all_levels[current_level].map,sizeof(current_map));
    found_keys = false;
    level_intro = true;
    level_intro_pos = 1.2f;
}

void renderTutorial()
{
    glClearColor(0, 0, 0, 1.0f);
    glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);

    if (touch_event)
    {
        game_screen = MAIN_MENU;
        touch_event = false;
    }
    drawObjectWithScale(TUTORIAL_CONTENTS, 0.0f, 0.0f, 1.0f);
}

void renderCredits()
{
    glClearColor(0, 0, 0, 1.0f);
    glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);

    if (touch_event)
    {
        game_screen = MAIN_MENU;
        touch_event = false;
    }
    drawObjectWithScale(CREDITS_CONTENTS, 0.0f, 0.0f, 1.0f);
}

void renderMainMenu()
{
    static int cycles = 0;
    cycles++;

    glClearColor(0, 0, 0, 1.0f);
    glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);

    if (isStartTouched())
    {
        restartGame();
        game_screen = GAME_PLAY;
    }
    else if (isTutorialTouched())
    {
        game_screen = TUTORIAL_SCREEN;
    }
    else if (isCreditsTouched())
    {
        game_screen = CREDITS_SCREEN;
    }
    touch_event = false;

    drawObjectWithScale(OPENING,0.0f,0.0f,1.0f);

    if ((cycles/20)%2)
    {
        drawObjectWithScale(START,0.0f,0.0f,0.4f);
    }

    drawObjectWithScale(TUTORIAL,-0.5f,-0.4f,0.4f);
    drawObjectWithScale(CREDITS,0.5f,-0.4f,0.4f);
}

void renderGame()
{
    static int cycles = 0;

    glClearColor(0, 0, 0, 1.0f);
    glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);

    killed_by_monster = false;
    killed_by_ghost = false;
    just_found_keys = false;

    int tx = 0, ty = 0;
    bool touch = getTouchedCell(tx, ty);

    if (isPausePlayTouched())
    {
        paused = !paused;
    }

    touch_event = false;

    if (!paused && !level_intro)
    {
        cycles++;

        for (int i = 0 ; i < all_levels[current_level].num_ghosts ; i++)
        {
            if (cycles%current_ghosts[i].ghost_cycle == 0)
                updateGhostPosition(&current_ghosts[i],cycles % 13 > 10);
        }

        /* If touched an arrow to move */
        if (touch && !fire_photon_torpedos && !diededed && !reached_goal)
        {
            if ((tx == bat_x+1 || tx == bat_x+2) && ty == bat_y && map(bat_x+1, bat_y) != 1)
            {
                /* RIGHT */
                bat_x++;
                last_bat_movement = 3;
                updateCharacterFrame(last_bat_movement);
            }
            else if ((tx == bat_x-1 || tx == bat_x-2) && ty == bat_y && map(bat_x-1, bat_y) != 1)
            {
                /* LEFT */
                bat_x--;
                last_bat_movement = 2;
                updateCharacterFrame(last_bat_movement);
            }
            else if (tx == bat_x && (ty == bat_y+1 || ty == bat_y+2) && map(bat_x, bat_y+1) != 1)
            {
                /* UP */
                bat_y++;
                last_bat_movement = 0;
                updateCharacterFrame(last_bat_movement);
            }
            else if (tx == bat_x && (ty == bat_y-1 || ty == bat_y-2) && map(bat_x, bat_y-1) != 1)
            {
                /* DOWN */
                bat_y--;
                last_bat_movement = 1;
                updateCharacterFrame(last_bat_movement);
            }
            else if (tx == bat_x && ty == bat_y)
            {
                /* SCAN */
                fire_photon_torpedos = true;
                wave_rad = 0.1;

                for (int i = 0 ; i < all_levels[current_level].num_ghosts ; i++)
                {
                    current_ghosts[i].ghost_goal_detected = true;
                    current_ghosts[i].ghost_goal_x = bat_x;
                    current_ghosts[i].ghost_goal_y = bat_y;
                }
            }
        }

        if (cycles % 3)
        {
            if (fire_photon_torpedos)
            {
                if (!stable_photon_torpedos)
                {
                    wave_rad += 0.02f;
                    if (wave_rad > 0.5f)
                    {
                        stablizing_counter = 13;
                        stable_photon_torpedos = true;
                    }
                }
                else
                {
                    stablizing_counter--;
                    if (stablizing_counter <= 0)
                    {
                        fire_photon_torpedos = false;
                        stable_photon_torpedos = false;
                        wave_rad = 0;
                    }
                }
            }
        }

        if (cycles % 10 == 0)
            updateCharacterFrame(last_bat_movement);

        /* If eaten by monster */
        if (map(bat_x, bat_y) == 2 && !diededed)
        {
            wave_rad = 0;
            fire_photon_torpedos = false;
            stable_photon_torpedos = false;
            last_bat_movement = 1;
            updateCharacterFrame(last_bat_movement);
            for (int i = 0 ; i < all_levels[current_level].num_ghosts ; i++) {
                current_ghosts[i].ghost_goal_detected = false;
            }
            diededed = true;
            diededed_counter = 25;
            killer_x = bat_x;
            killer_y = bat_y;
            killed_by_monster = true;
        }

        /* If ran into ghost */
        for (int i = 0 ; i < all_levels[current_level].num_ghosts ; i++)
        {
            if (abs(bat_x - current_ghosts[i].ghost_x) <= 1 && abs(bat_y - current_ghosts[i].ghost_y) <= 1 &&
                    !diededed && current_ghosts[i].ghost_goal_detected)
            {
                wave_rad = 0;
                fire_photon_torpedos = false;
                stable_photon_torpedos = false;
                last_bat_movement = 1;
                updateCharacterFrame(last_bat_movement);
                current_ghosts[i].ghost_goal_detected = false;
                diededed = true;
                diededed_counter = 25;
                killer_x = -1;
                killer_y = -1;
                killed_by_ghost = true;
            }
        }

        /* If found key */
        if (map(bat_x, bat_y) == 3)
        {
            found_keys = true;
            just_found_keys = true;
            takeKey(bat_x,bat_y);
        }

        if (diededed)
        {
            if (cycles % 5 == 0)
            {
                diededed_counter--;
                if (diededed_counter <= 0)
                {
                    diededed_counter = 0;
                    diededed = false;
                    bat_x = 0;
                    bat_y = 0;
                }
            }
        }
    }

    if (!reached_goal && (found_keys || !all_levels[current_level].has_key))
    {
        reached_goal = (bat_x == all_levels[current_level].goal_x) && (bat_y == all_levels[current_level].goal_y);
        if (reached_goal)
        {
            reached_goal_count = 30;
        }
    }

    if (reached_goal)
    {
        for (int i = 0 ; i < all_levels[current_level].num_ghosts ; i++)
        {
            current_ghosts[i].ghost_goal_detected = false;
        }
    }

    for (int x = 0 ; x < 10 ; x++)
    {
        for (int y = 0 ; y < 10 ; y++)
        {
            if (map(x,y) == 1)
            {
                drawObject(WALL,x,y,wave_rad);
            }
            else if (map(x,y) == 2)
            {
                float tmp_wave_rad = wave_rad;
                if (diededed && killer_x == x && killer_y == y)
                {
                    tmp_wave_rad = 10.0f;
                }
                drawObject(GROUND,x,y,wave_rad);
                drawObject(GameObject(EATER_1+(cycles/20)%2),x,y,tmp_wave_rad);
            }
            else if (map(x,y) == 0)
            {
                drawObject(GROUND,x,y,wave_rad);
            }
            if (map(x,y) == 3)
            {
                drawObject(GROUND,x,y,wave_rad);
                drawObject(KEY,x,y,wave_rad);
            }
        }
    }

    if (landscape)
    {
        float x1 = min_x - cell_x;
        float x2 = -x1 + cell_x*width_ratio;
        for (float y = -0.9f ; y < 1.0f ; y += cell_y)
        {
            drawObject(WALL,x1,y);
            drawObject(WALL,x2,y);
        }
    }
    else
    {
        float y1 = min_y - cell_y;
        float y2 = -min_y + cell_y;
        for (float x = -0.9f ; x < 1.0f ; x += cell_x)
        {
            drawObject(WALL,x,y1);
            drawObject(WALL,x,y2);
        }
    }

    drawObject(ENTERANCE, 0, 0, 10.0f);

    if (!all_levels[current_level].has_key || found_keys)
    {
        drawObject(EXIT_OPEN, all_levels[current_level].goal_x, all_levels[current_level].goal_y, 10.0f);
    }
    else
    {
        drawObject(EXIT, all_levels[current_level].goal_x, all_levels[current_level].goal_y, 10.0f);
    }

    /* Drawing Ghost */
    for (int i = 0 ; i < all_levels[current_level].num_ghosts ; i++)
    {
        drawObject(current_ghosts[i].ghost_frame, current_ghosts[i].ghost_x, current_ghosts[i].ghost_y, 10.0f);
    }

    /* Drawing Character */
    if (!diededed)
    {
        if (!reached_goal)
        {
            drawObject(character_frame,bat_x,bat_y,10.0f);
        }
        else
        {
            drawObject(BAT_FRONT_2,bat_x,bat_y,10.0f);
        }
    }
    else
    {
        drawObject(DEAD,bat_x,bat_y,10.0f);
    }

    if (!fire_photon_torpedos && !reached_goal && !diededed)
    {
        /* Drawing Arrows */
        if (bat_x+1 < 10)
            drawObject(ARROW_RIGHT,bat_x+1,bat_y, 10.0f);
        if (bat_x-1 >= 0)
            drawObject(ARROW_LEFT,bat_x-1,bat_y, 10.0f);
        if (bat_y+1 < 10)
            drawObject(ARROW_UP,bat_x,bat_y+1, 10.0f);
        if (bat_y-1 >= 0)
            drawObject(ARROW_DOWN,bat_x,bat_y-1, 10.0f);
    }

    if (!paused)
    {
        drawObject(PAUSE,pause_button_x,pause_button_y);
        if (reached_goal)
        {
            drawObjectWithScale(NICE, 0.0f, 0.0f, 0.65f);
        }
    }
    else
    {
        drawObject(PLAY,pause_button_x,pause_button_y);
        drawObjectWithScale(PAUSED, 0.0f, 0.0f, 0.65f);
    }

    if (level_intro)
    {
        WriteText(all_levels[current_level].message, 0.0f, -0.8f, 0.025f, true);

        drawObjectWithScale(GameObject(LV1+current_level),level_intro_pos,0.0f,0.4f);
        level_intro_pos -= 0.013f;
        if (level_intro_pos < -1.5f)
        {
            level_intro = false;
        }
    }

    if (reached_goal && cycles%5 == 0)
    {
        reached_goal_count--;
        if (reached_goal_count <= 0)
        {
            if (current_level >= NUM_LEVELS-1)
            {
                game_screen = MAIN_MENU;
            }
            else
            {
                current_level++;
                restartGame();
            }
        }
    }
}

extern "C" {
    JNIEXPORT void JNICALL Java_com_android_slepi_GL2JNILib_init(JNIEnv * env, jobject obj,  jint width, jint height);
    JNIEXPORT void JNICALL Java_com_android_slepi_GL2JNILib_step(JNIEnv * env, jobject obj);
    JNIEXPORT void JNICALL Java_com_android_slepi_GL2JNILib_onTouchEvent(JNIEnv * env, jobject obj, jfloat x, jfloat y);
    JNIEXPORT void JNICALL Java_com_android_slepi_GL2JNILib_loadImage(JNIEnv * env, jobject obj, jint width, jint height, jintArray buffer);
    JNIEXPORT int JNICALL Java_com_android_slepi_GL2JNILib_getAudioMask(JNIEnv * env, jobject obj);
    JNIEXPORT void JNICALL Java_com_android_slepi_GL2JNILib_pause(JNIEnv * env, jobject obj);
    JNIEXPORT void JNICALL Java_com_android_slepi_GL2JNILib_resume(JNIEnv * env, jobject obj);
};

JNIEXPORT void JNICALL Java_com_android_slepi_GL2JNILib_init(JNIEnv * env, jobject obj,  jint width, jint height)
{
    setupGraphics(width, height);
    InitializeLevels();
}

JNIEXPORT void JNICALL Java_com_android_slepi_GL2JNILib_step(JNIEnv * env, jobject obj)
{
    if (game_screen == MAIN_MENU)
    {
        renderMainMenu();
    }
    else if (game_screen == GAME_PLAY)
    {
        renderGame();
    }
    else if (game_screen == TUTORIAL_SCREEN)
    {
        renderTutorial();
    }
    else if (game_screen == CREDITS_SCREEN)
    {
        renderCredits();
    }
}

JNIEXPORT void JNICALL Java_com_android_slepi_GL2JNILib_onTouchEvent(JNIEnv * env, jobject obj, jfloat x, jfloat y)
{
    touch_x = x;
    touch_y = y;
    touch_event = true;
}

JNIEXPORT void JNICALL Java_com_android_slepi_GL2JNILib_loadImage(JNIEnv * env, jobject obj, jint width, jint height, jintArray buffer)
{
    jboolean isCopy = 0;
    jint* data = env->GetIntArrayElements(buffer, &isCopy);

    int size = width*height*4;

    char* src = (char*)data;
    char* dst = new char[size];

    memcpy(dst,src,(size_t)size);

    int index = 0;
    for (int y = 0 ; y < height ; y++)
    {
        for (int x = 0 ; x < width ; x++)
        {
            char tmp = dst[index];
            dst[index] = dst[index+2];
            dst[index+2] = tmp;
            index += 4;
        }
    }

    gTextureList[gNumTextureList].format = GL_RGBA;
    gTextureList[gNumTextureList].data = dst;
    gTextureList[gNumTextureList].width = width;
    gTextureList[gNumTextureList].height = height;
    gNumTextureList++;
}

JNIEXPORT int JNICALL Java_com_android_slepi_GL2JNILib_getAudioMask(JNIEnv * env, jobject obj)
{
    int mask = 0;

    mask |= (game_screen == GAME_PLAY ? 1 : 0) << 0;
    mask |= (fire_photon_torpedos ? 1 : 0) << 1;
    mask |= (killed_by_monster ? 1 : 0) << 2;
    mask |= (killed_by_ghost ? 1 : 0) << 3;
    mask |= (just_found_keys ? 1 : 0) << 4;
    mask |= (reached_goal_count == 20 ? 1 : 0) << 5;
    mask |= (reached_goal_count == 29 ? 1 : 0) << 6;

    return mask;
}

JNIEXPORT void JNICALL Java_com_android_slepi_GL2JNILib_pause(JNIEnv * env, jobject obj)
{

}

JNIEXPORT void JNICALL Java_com_android_slepi_GL2JNILib_resume(JNIEnv * env, jobject obj)
{

}
